/* dnsq.c:
 * Dns query.
 * This builds on Cygwin and solaris, haven't tested it elsewhere.
 *
 * On sunglare build this with: 
 *  gcc -o dnsq dnsq.c -lnsl -lsocket -lresolv
 * on cygwin:
 *  gcc -o dnsq dnsq.c -liphlpapi
 */
#if defined (__CYGWIN32__)
#include <windows.h>
#include <iphlpapi.h>
#else
#define INET_ADDRSTRLEN 64
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <resolv.h>

#ifndef INADDR_NONE
#define INADDR_NONE 0xffffffff
#endif

#define SOCKET_INVALID(s)	(s < 0)

#define DNS_PORT	 53

struct sockaddr_in targetAddr;

int getResponse(int sockfd, char *msgbuf, int mbsiz, int msecwait, int tcp);

char *errmsg[] = {
	"Usage: dnsq [options]",
	" Options:",
	"  -v     verbose mode",
	(char *)0
};

void
usage()
{
	int i = 0;

	while(errmsg[i])
		printf("%s\n",errmsg[i++]);
	exit(1);
}

void
err(char *msg)
{
	if (msg)
		fprintf(stderr,"%s, ",msg);
	perror("");
	exit(-1);
}

int
main(int argc,char **argv)
{
	uint32_t inaddr;
	char	*gatewayIP;
	char	ibuf[INET_ADDRSTRLEN];
	struct	hostent	*hp, host_info;
	int		i, rc, opt, pktsiz, len, sfd;
	int		verbose = 1;
	int		useTCP = 0;		/* never changes, see notes above */
	char	natpmprqst[256], natpmpresp[256];

	if (argc != 1)
		usage();
		
	while((opt=getopt(argc,argv,"v")) != EOF) { 
		switch(opt) {
		case 'v':
			verbose = 1;
			break;
		}
	}

	if (res_init() == -1) {
		fprintf(stderr,"res_init() failed\n");
		exit(1);
	}

	if (_res.nscount <= 0) {
#if defined (__CYGWIN32__)
		ULONG buflen = 0;
		FIXED_INFO *buf = NULL;
		if (GetNetworkParams(NULL, &buflen) == ERROR_BUFFER_OVERFLOW)
			buf = malloc(buflen);
		if (buf && GetNetworkParams(buf, &buflen) == NO_ERROR)
		{
			_res.nscount = 1;
			_res.nsaddr_list[0].sin_family = AF_INET;
			_res.nsaddr_list[0].sin_addr.s_addr = inet_addr(buf->DnsServerList.IpAddress.String);
			_res.nsaddr_list[0].sin_port = htons(53);
		}
		free(buf); 

		for(i=0;i<_res.nscount;i++) {
			fprintf(stderr, "dnssrvr: %s\n",
				inet_ntop(AF_INET, &_res.nsaddr_list[i].sin_addr, ibuf, sizeof ibuf)); 
		}
#else
		fprintf(stderr,"nscount = %d\n",_res.nscount);
#endif
	}
	else {
		for(i=0;i<_res.nscount;i++) {
			fprintf(stderr, "dnssrvr: %s\n",
				inet_ntop(AF_INET, &_res.nsaddr_list[i].sin_addr, ibuf, sizeof ibuf)); 
		}
	}
	exit(0);

	gatewayIP = "192.168.1.1";

	if (argc == optind+3) 
		gatewayIP = argv[optind+2];
	else if (argc != optind+2)
		usage();
	
	/* Accept target name as string or internet dotted-decimal address:
	 */
	memset((char *)&targetAddr,0,sizeof(struct sockaddr));
	if ((inaddr = inet_addr(gatewayIP)) != INADDR_NONE) {
		memcpy((char *)&targetAddr.sin_addr,(char *)&inaddr,sizeof(inaddr));
		host_info.h_name = NULL;
	}
	else {
		hp = gethostbyname(gatewayIP);
		if (hp == NULL)
			err("gethostbyname failed");
		host_info = *hp;
		memcpy((char *)&targetAddr.sin_addr,hp->h_addr,hp->h_length);
	}

	if (useTCP)
		sfd = socket(AF_INET,SOCK_STREAM,0);
	else
		sfd = socket(AF_INET,SOCK_DGRAM,0);

	if (SOCKET_INVALID(sfd))
		err("socket failed");

	targetAddr.sin_family = AF_INET;
	targetAddr.sin_port = htons(DNS_PORT);

	if (useTCP) {
		rc = connect(sfd,(struct sockaddr *)&targetAddr,
				sizeof(struct sockaddr));
		if (rc < 0) {
			perror("connect");
			exit(1);
		}
	}

	
	/* Build the NAT-PMP public address request:
	 * ---------------------------
	 * |   BYTE-0   |   BYTE-1   |
	 * ---------------------------
	 * | Version=0  |  OPCode=0  |
	 * ---------------------------
	 */
	natpmprqst[0] = 0x00;	/* Version */
	natpmprqst[1] = 0x00;	/* Opcode */
	
	pktsiz = 2;
	
	if (useTCP) {
		rc = send(sfd,natpmprqst,pktsiz,0);
	}
	else {
		rc = sendto(sfd,natpmprqst,pktsiz,0,
	   	 (struct sockaddr *)&targetAddr,sizeof(targetAddr));
	}

	if (rc < 0) {
		close(sfd);
		err("send/sendto failed");
	}

	/* If a response is received, it should be formatted as follows:
	 * -----------------------------------------------------
	 * |   BYTE-0   |   BYTE-1   |   BYTE-2   |   BYTE-3   |
	 * -----------------------------------------------------
	 * | Version=0  | OPCode=128 |      Result Code        |
	 * -----------------------------------------------------
	 * |   BYTE-4   |   BYTE-5   |   BYTE-6   |   BYTE-7   |
	 * -----------------------------------------------------
	 * | Seconds since port mapping table was initialized  |
	 * -----------------------------------------------------
	 * |   BYTE-8   |   BYTE-9   |   BYTE-10  |   BYTE-11  |
	 * -----------------------------------------------------
	 * |                  Public IP address                |
	 * -----------------------------------------------------
	 */

	len = getResponse(sfd,natpmpresp,sizeof(natpmpresp), 10000, useTCP);
	 
	if (len <= 0)
		exit(1);

	exit(0);
}

int
getResponse(int sockfd, char *msgbuf, int mbsiz, int msecwait, int tcp)
{
	int	msglen, rcvlen;
	fd_set	fd_read;
	struct	timeval tval;

	tval.tv_sec = 0;
	tval.tv_usec = msecwait * 1000;

	FD_ZERO(&fd_read);
	FD_SET(sockfd,&fd_read);

	/* Wait for incoming message: */
	if (select(sockfd+1,&fd_read,0,0,&tval) != 1) {
		return(-1);
	}

	if (tcp) {
		rcvlen = recv(sockfd,msgbuf,mbsiz,0);
	}
	else {
		msglen = sizeof(struct sockaddr);
		rcvlen = recvfrom(sockfd,msgbuf,mbsiz,0,
			(struct sockaddr *)&targetAddr,&msglen);
	}
	if (rcvlen == 0) {
		fprintf(stderr,"Connection closed\n");
	}
	else if (rcvlen < 0) {
		if (tcp)
			perror("recv");
		else
			perror("recvfrom");
	}

	return(rcvlen);
}

